class ShippingMethodFactory(object):
    @classmethod
    def shipping_method(klass, shipping_preference, shipping_methods_response):
        if hasattr(klass, shipping_preference):
            return getattr(klass, shipping_preference)(shipping_methods_response)
        else:
            raise TypeError("The shipping preference '%s' is not supported." % shipping_preference)

    @classmethod
    def cheapest(klass, shipping_methods_response):
        minimum = (None, None)
        for method in shipping_methods_response["shipping_methods"]:
            if minimum[0] == None or method["price"] < minimum[0]:
                minimum = (method["price"], method["shipping_method_id"])

        if minimum[0] == None:
            return shipping_methods_response["shipping_methods"][0]["shipping_method_id"]
        else:
            return minimum[1]

    @classmethod
    def second(klass, shipping_methods_response):
        possible_methods = ['second','std-n-us','std-us', 'Std Cont US Street Addr', 'Std US Dom Plus', 'second-non48', 'std-n-us-non48', 'std-us-non48']

        for shipping_id in possible_methods:
            for method in shipping_methods_response["shipping_methods"]:
                if method["shipping_method_id"] == shipping_id:
                    return shipping_id
        raise Exception("The response doesn't have any of the shipping methods you requested")

    @classmethod
    def standard(klass, shipping_methods_response):
        possible_methods = ['std-n-us','std-us', 'Std Cont US Street Addr', 'Std US Dom Plus', 'std-n-us-non48', 'std-us-non48']

        for shipping_id in possible_methods:
            for method in shipping_methods_response["shipping_methods"]:
                if method["shipping_method_id"] == shipping_id:
                    return shipping_id
        raise Exception("The response doesn't have any of the shipping methods you requested")



import requests
import time
import json

class MaximumRequestRetriesExceeded(Exception):
    pass

class ZincTimeoutError(Exception):
    pass

class ZincError(Exception):
    pass

class ZincAsyncRequest(object):
    def __init__(self, url, request_id, abstract_processor, start_time):
        self.url = url
        self.request_id = request_id
        self.abstract_processor = abstract_processor
        self.start_time = start_time

    def finished(self):
        self.abstract_processor.finished(self.url, self.request_id)

    def get_response(self):
        self.abstract_processor.wait_for_response(self.url, self.request_id, self.start_time)

class ZincRequestProcessor(object):
    """Processor for Zinc API requests.

    Simple wrapper that instantiates and calls the ZincAbstractProcessor.
    """
    @classmethod
    def process(klass, request_type, payload,
            zinc_base_url="https://api.zinc.io/v0",
            polling_interval=1.0):
        processor = ZincAbstractProcessor()
        return processor.post_request(payload, request_type)

    @classmethod
    def process_async(klass, request_type, payload,
            zinc_base_url="https://api.zinc.io/v0"):
        processor = ZincAbstractProcessor()
        return processor.post_request(payload, request_type)

class ZincAbstractProcessor(object):
    def __init__(self, zinc_base_url="https://api.zinc.io/v0", 
            polling_interval = 1.0,
            request_timeout = 180,
            get_request_timeout = 5.0,
            post_request_timeout = 10.0,
            get_request_retries = 3):
        self.zinc_base_url = zinc_base_url
        self.polling_interval = polling_interval
        self.request_timeout = request_timeout
        self.get_request_timeout = get_request_timeout
        self.post_request_timeout = post_request_timeout
        self.get_request_retries = get_request_retries

    def current_url(self, url_stub = None):
        if url_stub:
            return self.zinc_base_url + "/" + url_stub
        else:
            return self.zinc_base_url

    def post_request(self, payload, url_stub, async=False):
        start_time = time.time()
        result = requests.post(self.current_url(url_stub), data=json.dumps(payload), timeout=self.post_request_timeout)
        request_id = result.json()["request_id"]
        if async:
            return ZincAsyncRequest(self.current_url(url_stub), request_id, self, start_time)
        else:
            return self.wait_for_response(self.current_url(url_stub), request_id, start_time)

    def wait_for_response(self, url, request_id, start_time=None):
        if start_time != None and time.time() - start_time > self.request_timeout:
            raise ZincTimeoutError("The request '%s' timed out after '%s' seconds" %
                    (request_id, time.time() - start_time))
        time.sleep(self.polling_interval)
        potential_result = self.finished(url, request_id)
        if potential_result:
            end_time = time.time()
            return potential_result
        else:
            return self.wait_for_response(url, request_id, start_time)

    def finished(self, url, request_id):
        result = self.make_request(url + "/" + request_id, self.get_request_retries)
        result_json = result.json()
        if result_json["_type"] == "error" and result_json["code"] == "request_processing":
            return False
        elif result_json["_type"] == "error":
            raise ZincError(result)
        else:
            return result_json

    def make_request(self, full_url, retries):
        try:
            return requests.get(full_url, timeout=self.get_request_timeout)
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
            if retries > 0:
                return self.make_request(full_url, retries-1)
            else:
                raise MaximumRequestRetriesExceeded("Maximum number of request retries exceeded for connecting to the Zinc API")




from zinc_request_processor import ZincAbstractProcessor, ZincTimeoutError, \
    MaximumRequestRetriesExceeded, ZincError
from shipping_method_factory import ShippingMethodFactory
import json

class ZincSimpleOrder(object):
    def __init__(self,
            zinc_base_url="https://api.zinc.io/v0",
            max_tries = 3,
            polling_interval= 1.0,
            request_timeout = 180,
            get_request_timeout = 5.0,
            post_request_timeout = 10.0,
            get_request_retries = 3):
        self.shipping_method_factory = ShippingMethodFactory
        self.max_tries = max_tries
        self.processor = ZincAbstractProcessor(zinc_base_url=zinc_base_url,
                polling_interval=polling_interval,
                request_timeout=request_timeout,
                get_request_timeout=get_request_timeout,
                post_request_timeout=post_request_timeout,
                get_request_retries=get_request_retries)

    def process(self, order_details):
        return self._process_helper(order_details, self.max_tries)

    def process_json(self, json_order_details, multiple=False):
        if multiple:
            return self.process_multiple(json.loads(json_order_details))
        else:
            return self.process(json.loads(json_order_details))

    def process_file(self, filename, multiple=False):
        with open(filename, 'rb') as f:
            return self.process_json(f.read(), multiple)

    def _process_helper(self, order_details, tries):
        if tries > 0:
            try:
                shipping_methods_response = self.processor.post_request(
                        self._shipping_methods_details(order_details), "shipping_methods")
                store_card_response = self.processor.post_request(
                        self._store_card_details(order_details), "store_card")
                review_order_response = self.processor.post_request(
                        self._review_order_details(order_details, shipping_methods_response,
                            store_card_response), "review_order")
            except (ZincTimeoutError, MaximumRequestRetriesExceeded, ZincError):
                return self._process_helper(order_details, tries-1)

            # We don't want to retry on place order requests, in case we accidentally
            # place multiple orders.
            place_order_response = self.processor.post_request(
                    self._place_order_details(order_details, review_order_response),
                    "place_order")
            return place_order_response
        else:
            raise Exception("Maximum tries exceeded: unable to place Zinc order")

    def _shipping_methods_details(self, order_details):
        return {
                "client_token": order_details["client_token"],
                "retailer": order_details["retailer"],
                "shipping_address": order_details["shipping_address"],
                "products": self._products(order_details)
                }

    def _store_card_details(self, order_details):
        return {
                "client_token": order_details["client_token"],
                "billing_address": order_details["billing_address"],
                "number": order_details["payment_method"]["number"],
                "expiration_month": order_details["payment_method"]["expiration_month"],
                "expiration_year": order_details["payment_method"]["expiration_year"]
                }

    def _review_order_details(self, order_details, shipping_methods_response,
            store_card_response):
        return {
                "client_token": order_details["client_token"],
                "retailer": order_details["retailer"],
                "products": self._products(order_details),
                "shipping_address": order_details["shipping_address"],
                "is_gift": order_details["is_gift"],
                "shipping_method_id": self._shipping_method_id(order_details,
                    shipping_methods_response),
                "payment_method": {
                    "security_code": order_details["payment_method"]["security_code"],
                    "cc_token": store_card_response["cc_token"]
                    },
                "customer_email": order_details["customer_email"]
                }

    def _place_order_details(self, order_details, review_order_response):
        return {
                "client_token": order_details["client_token"],
                "place_order_key": review_order_response["place_order_key"]
                }

    def _products(self, order_details):
        return [{
            "product_id": order_details["product_id"],
            "quantity": order_details["quantity"]
            }]


    def _shipping_method_id(self, order_details, shipping_methods_response):
        return self.shipping_method_factory.shipping_method(
                order_details["shipping_preference"], shipping_methods_response)
